﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace RoboRace
{
    public class GameLogicException : Exception
    { }

    public class GameLogic
    {
        public static bool checkValidity(GameState state, string moves, bool previewMode)
        {
            int[] availableCounts = new int[256];
            int[] moveCounts = new int[256];
            for (int i = 0; i < 256; i++) { availableCounts[i] = 0; moveCounts[i] = 0; }
            for (int i = 0; i < state.possibleMoves.Count; i++) { availableCounts[(int)state.possibleMoves[i]]++; }
            for (int i = 0; i < moves.Length; i++) { moveCounts[moves[i]]++; }
            for (int i = 0; i < moves.Length; i++) {
                if (moveCounts[moves[i]] > availableCounts[moves[i]]) return false; 
            }

            try
            {
                GameState temp = patchState(state, moves, false);
                if (!previewMode && moves.Length != 5 && (temp.robotX != temp.endX || temp.robotY != temp.endY))
                {
                    return false;
                }
            }
            catch (GameLogicException ex)
            {
                return false;
            }

            return true;
        }

        public static GameState patchState(GameState s, string moves, bool solving)
        {
            GameState state = new GameState(s);
            moves = moves.Replace("2", "91");
            moves = moves.Replace("3", "991");
            int[,] d = {{0,-1}, {1, 0}, {0, 1}, {-1, 0}};
            char c;
            for (int i = 0; i < moves.Length; i++ )
            {
                if (state.robotX == state.endX && state.robotY == state.endY && !solving) continue;
                c = moves[i];
                int nx, ny;
                if (c == 'R') state.robotRotation = (state.robotRotation + 1) % 4;
                else if (c == 'L') state.robotRotation = (state.robotRotation + 3) % 4;
                else if (c == 'Z')
                {
                    nx = state.robotX - d[state.robotRotation, 0];
                    ny = state.robotY - d[state.robotRotation, 1];
                    if (nx < 0 || nx >= 20 || ny < 0 || ny >= 20 || state.map[ny][nx] == '#') throw new GameLogicException();
                    if (state.map[ny][nx] != 'X') { state.robotX = nx; state.robotY = ny; }
                }
                else if (c == '1' || c == '9')
                {
                    nx = state.robotX + d[state.robotRotation, 0];
                    ny = state.robotY + d[state.robotRotation, 1];
                    if (nx < 0 || nx >= 20 || ny < 0 || ny >= 20 || state.map[ny][nx] == '#') throw new GameLogicException();
                    if (state.map[ny][nx] != 'X') { state.robotX = nx; state.robotY = ny; }
                }

                int rotIndex = -1;
                switch (state.map[state.robotY][state.robotX])
                {
                    case '^':
                        rotIndex = 0;
                        break;
                    case '>':
                        rotIndex = 1;
                        break;
                    case 'v':
                        rotIndex = 2;
                        break;
                    case '<':
                        rotIndex = 3;
                        break;
                }

                if (rotIndex != -1 && c != '9')
                {
                    nx = state.robotX + d[rotIndex, 0];
                    ny = state.robotY + d[rotIndex, 1];
                    if (nx < 0 || nx >= 20 || ny < 0 || ny >= 20 || state.map[ny][nx] == '#') throw new GameLogicException();
                    if (state.map[ny][nx] != 'X') { state.robotX = nx; state.robotY = ny; }
                }
            }

            return state;
        }

        public static string findMoves(GameState start, int targetX, int targetY, int targetRotation)
        {
            Queue<GameState> q = new Queue<GameState>();
            q.Enqueue(start);
            while (q.Count > 0)
            {
                GameState current = q.Dequeue();
                if (current.doneMoves.Count == 5)
                {
                    if (current.robotX == targetX && current.robotY == targetY && current.robotRotation == targetRotation)
                    {
                        StringBuilder sb = new StringBuilder();
                        foreach (char c in current.doneMoves)
                        {
                            sb.Append(c);
                        }
                        return sb.ToString();
                    }
                }
                else
                {
                    foreach (char c in current.possibleMoves)
                    {
                        try
                        {
                            GameState newState = GameLogic.patchState(current, c.ToString(), true);
                            List<char> npossible = current.possibleMoves.ToArray().ToList<char>();
                            npossible.Remove(c);
                            newState.possibleMoves = npossible;
                            newState.doneMoves.Add(c);
                            q.Enqueue(newState);
                        }
                        catch (GameLogicException ex)
                        {
                        }
                    }
                }
            }

            return "";
        }

        public static void Solve(Connector connector)
        {
            GameState start = connector.getCurrentState();
            if (start.robotX == start.endX && start.robotY == start.endY) return;
            Queue<GameState> q = new Queue<GameState>();
            q.Enqueue(start);
            GameState next;
            string choosenMove = "";
            bool solved = false;
            while (q.Count > 0)
            {
                GameState current = q.Dequeue();
                if (current.doneMoves.Count == 5)
                {
                    StringBuilder sb = new StringBuilder();
                    foreach (char c in current.doneMoves)
                    {
                        sb.Append(c);
                    }
                    if (current.robotX == current.endX && current.robotY == current.endY)
                    {
                        connector.sendMoves(sb.ToString());
                        solved = true;
                        return;
                    }
                    if (!solved && GameLogic.checkValidity(connector.getCurrentState(), sb.ToString(), false))
                    {
                        choosenMove = sb.ToString();
                    }
                }
                else
                {
                    foreach (char c in current.possibleMoves)
                    {
                        try
                        {
                            GameState newState = GameLogic.patchState(current, c.ToString(), false);
                            List<char> npossible = current.possibleMoves.ToArray().ToList<char>();
                            npossible.Remove(c);
                            newState.possibleMoves = npossible;
                            newState.doneMoves.Add(c);
                            q.Enqueue(newState);
                        }
                        catch (GameLogicException ex)
                        {
                        }
                    }
                }
            }
            q.Clear();
            connector.sendMoves(choosenMove);

            Solve(connector);
        }
    }
}
